/* auto-generated by util/mkbuildinf.pl for crypto/cversion.c */
#define CFLAGS cflags
/*
 * Generate CFLAGS as an array of individual characters. This is a
 * workaround for the situation where CFLAGS gets too long for a C90 string
 * literal
 */
static const char cflags[] = {
    'c','o','m','p','i','l','e','r',':',' ','c','c','\0'
};
#define PLATFORM "platform: "
#define DATE "built on: Tue Apr  3 00:38:58 2018"
